#!/usr/bin/python

import serial
import sys
import glob
import wave
import struct
import time
import re
import os
import time

DATA_SIZE = 64 # max = 64
DATA_CHUNK = 16*32
FREQ_BINS = 16
timer = 23
num_chunks = 64

def serial_ports():
  """Lists serial ports
  Raises:
  EnvironmentError:
      On unsupported or unknown platforms
  Returns:
      A list of available serial ports
  """
  if sys.platform.startswith('win'):
    ports = ['COM' + str(i + 1) for i in range(256)]
  elif sys.platform.startswith('linux') or sys.platform.startswith('cygwin'):
      # this is to exclude your current terminal "/dev/tty"
    ports = glob.glob('/dev/tty[A-Za-z]*')
  elif sys.platform.startswith('darwin'):
    ports = glob.glob('/dev/tty.*')
  else:
    raise EnvironmentError('Unsupported platform')
  result = []
  for port in ports:
    try:
      s = serial.Serial(port)
      s.close()
      result.append(port)
    except (OSError, serial.SerialException):
      pass
  return result

def serial_write(ser, data):
  if sys.platform.startswith('win'):
    ser.write([data,])
  else:
    ser.write(str(data).encode())

def write_cos_table(ser):
  f = open('cos_table.txt', 'r')
  ser.write("\ncos_table\n".encode())
  line = ser.readline().decode()
  while(line != "Start receiving cos_table\n"):
    line = ser.readline().decode()
  print("MSP: " + line.rstrip('\n'))
  for i in range(FREQ_BINS*FREQ_BINS):
    ser.write(f.readline().encode())
  line = ser.readline().decode()
  while(line != "Done receiving cos_table\n"):
    line = ser.readline().decode()
  print("Done sending cos_table\n")
  f.close()

def start_decode(ser):
  start = time.time()
  buff = []
  f = open('dct_devil.txt', 'r')
  for i in range(num_chunks):
    print("Starting segment " + str(i) + "...")
    write_data_in(ser, f)
    ser.close()
    time.sleep(1+(int)(timer/1000))
    ser.open()
    reconnect(ser)
    read_data_out(ser, buff)
    print("Done with segment " + str(i) + ".\n")
  print("Done with processing.")
  f.close()
  save_wav(buff, 'result.wav')
  save_txt(buff, 'result.txt')
  print("Runtime: %.2f secs\n" % float(time.time()-start))
  return 0

def reconnect(ser):
  line = ser.readline().decode()
  while(not re.search("A", line)):
    line = ser.readline().decode()
  ser.write("\nB\n".encode())

def write_data_in(ser, f):
  ser.write("\ndata_in\n".encode())
  line = ser.readline().decode()
  while(line != "Start receiving data\n"):
    line = ser.readline().decode()
  for i in range(DATA_CHUNK):
    ser.write((str(float(f.readline())) + "\n").encode())

def read_data_out(ser, buff):
  line = ser.readline().decode()
  while(line != "Sending data_out\n"):
    line = ser.readline().decode()
  line = ser.readline().decode()
  line = ser.readline().decode()
  for i in range(DATA_CHUNK):
    buff.append((float)(line.rstrip('\n')))
    line = ser.readline().decode()
  print("Finished decoding " + str((int)(line.rstrip('\n'))) + " frequency bins.")
  line = ser.readline().decode()

def save_txt(buff, txtfile):
  print("Writing to " + txtfile + "...")
  if os.path.isfile(txtfile):
    os.remove(txtfile)
  text_file = open(txtfile, 'w')
  text_file.write('\n'.join([str(x) for x in buff]))
  text_file.close()

def save_wav(buff, wavfile):
  print("Writing to " + wavfile + "...")
  if os.path.isfile(wavfile):
    os.remove(wavfile)
  data = []
  for line in buff:
    d = max(min(int(float(line)*32767), 32767), -32767)
    data.append(struct.pack('h',d))
  w = wave.open(wavfile, 'wb')
  w.setparams((2, 2, 22050, len(data), 'NONE', 'not compressed'))
  for d in data:
    w.writeframes(d)
    w.writeframes(d)
  w.close()

def set_clock(ser, freq, portNo):
  ser.write(("\nset_clock\n" + str(freq) + "\n").encode())
  return

def show_timer(ser):
  ser.write(("\nshow_timer\n").encode())

def set_timer(ser, new_timer):
  print("Setting timer to " + new_timer + "ms\n")
  ser.write(("\nset_timer\n" + new_timer + "\n").encode())
  timer = new_timer

def print_menu():
  print('Clock is output in pin 2.2')
  print('Entering terminal. Usage:')
  print('    "g" or "go" to start computation')
  print('    "sl #" or "set_length #" to set number of data chunks to stream (max: ' + str(DATA_SIZE) + ')')
  print('    "sc #" or "set_clock #" to set clock to # kHz')
  print('    "t" or "time" to show timer duration in pin 2.0')
  print('    "st #" or "set_timer #" to set timer to # ms (1-1999ms, 0 gives ~4s)')
  print('    "cos" or "cos_table" to resend cosine table')
  print('    "h" or "help" to print this menu')
  print('    "q" or "quit" to quit terminal')
  print()

# MAIN FUNCTION - terminal functionality

def run():
  global num_chunks

  print("EE16B DSP Lab")
  print("Press enter to begin.")

  ports = serial_ports()
  if ports:
    print("Available serial ports:")
    for (i,p) in enumerate(ports):
      print("%d) %s"%(i+1,p))
  else:
    print("No ports available. Check serial connection and try again.")
    print("Exiting...")
    return

  portNo = input("Select the port to use: ")
  ser = serial.Serial(ports[int(portNo)-1])
  ser.baudrate=38400 # default 25MHz
  ser.timeout=1
  ser.write("5".encode())
  ser.readline()
  ser.timeout=5

  write_cos_table(ser)

  print_menu()

  while(True):
    user_input = input('>>> ')
    print()
    if (user_input in ['g', 'go']):
      start_decode(ser)
    elif (re.match("sl", user_input) or re.match("set_length", user_input)):
      new_chunks = re.search("\d+", user_input);
      if (new_chunks):
        num_chunks = int(new_chunks.group(0))
        if (num_chunks > DATA_SIZE):
          print("Maximum length is " + str(DATA_SIZE) + " segments; set to 64 segments.")
          num_chunks = DATA_SIZE
        print("Number of data chunks set to " + str(num_chunks))
      else:
        print("Please add number of data chunks as argument.\n")
    elif (re.match("sc", user_input) or re.match("set_clock", user_input)):
      newclk = re.search("\d+", user_input);
      if (newclk):
        newclk = int(newclk.group(0))
        set_clock(ser, (int)(newclk), portNo)
        write_cos_table(ser)
      else:
        print("Please add clock frequency in kHz as argument.\n")
    elif (user_input in ['t', 'timer']):
      show_timer(ser)
    elif (re.match("st", user_input) or re.match("set_timer", user_input)):
      if(re.search("\d+", user_input)):
        set_timer(ser, re.search("\d+", user_input).group(0))
      else:
        print("Please add time in ms as argument.\n")
    elif (user_input in ['cos', 'cos_table']):
      write_cos_table(ser)
    elif (user_input in ['h', 'help']):
      print_menu()
    elif (user_input in ['q', 'quit']):
      ser.close()
      print("Bye!\n")
      return
    else:
      print("Unrecognized command: " + user_input + "\n")
